package edu.thi.phga.aparapi_test;

import com.aparapi.Kernel;
import com.aparapi.Range;

public class OpenCLPrefix2 {

  /**
     Parallel Prefix implementation with 2 arrays instead
     of one array twice the size of the requested input.

     Accumulated execution time was equally good.
     Works up to 1 << 28 before OutOfMemoryError

     Rewrite of the original idea by: Prof. Dr. Schmidt <Ulrich.Schmidt@thi.de>
  */
  public static void start(int[] a) {
    int N = a.length;
    int b[] = new int[N];
    java.util.Arrays.fill(b, 1);
    Range r = Range.create(App.device, N, 4);
    OpenCLPrefix2Kernel k = new OpenCLPrefix2Kernel(a, b);
    k.setExplicit(true);
    k.setStepSize(1);
    k.execute(r);
    k.toggleIndex();
    double ct = k.getConversionTime();
    // Uncomment to get Debug output (Try with 1 << 3)
    // k.get(a);
    // k.get(b);
    // System.out.println(k.getIndex());
    // for (int i = 0; i < b.length; i++) {
    //   System.out.printf("a[%d] = %d\n", i, a[i]);
    // }
    // for (int i = 0; i < b.length; i++) {
    //   System.out.printf("b[%d] = %d\n", i, b[i]);
    // }

    for (int n = 2; n < N; n *= 2) {
      k.setStepSize(n);
      k.execute(r);
      k.toggleIndex();
      // Uncomment to get Debug output (Try with 1 << 3)
      // k.get(a);
      // k.get(b);
      // System.out.println(k.getIndex());
      // for (int i = 0; i < b.length; i++) {
      //   System.out.printf("a[%d] = %d\n", i, a[i]);
      // }
      // for (int i = 0; i < b.length; i++) {
      //   System.out.printf("b[%d] = %d\n", i, b[i]);
      // }
    }

    double et = k.getAccumulatedExecutionTime() - ct;

    // Last write was to A
    if (k.getIndex() > 0) {
      k.get(a);
      // Last write was to B
    } else {
      k.get(b);
      a = b;
    }
    System.out.printf("P2 - GOT: %d, %d, %d, TIME: %.2f ms\n",
      a[N - 3], a[N - 2], a[N - 1], et);

  }

  private static class OpenCLPrefix2Kernel extends Kernel {
    private int stepSize;
    private int index = 1;
    private int[] a, b;

    public OpenCLPrefix2Kernel(int[] a, int[] b) {
      this.a = a;
      this.b = b;
    }

    public void setStepSize(int s) {
      this.stepSize = s;
    }

    public void toggleIndex() {
      index = -index;
    }

    public int getIndex() {
      return index;
    }

    @Override
    public void run() {
      int i = getGlobalId();
      // A -> B
      if (index > 0) {
        if (i < stepSize) {
          b[i] = a[i];
        } else {
          b[i] = a[i] + a[i - stepSize];
        }
        // B -> A
      } else {
        if (i < stepSize) {
          a[i] = b[i];
        } else {
          a[i] = b[i] + b[i - stepSize];
        }
      }
    }
  }
}