package edu.thi.phga.aparapi_test;

import com.aparapi.Kernel;

public class OpenCLReduce5Kernel extends Kernel {
  private int[] a, result, summe;
  private int step = 1;
  @Local
  private int[] scratch;

  public OpenCLReduce5Kernel(int[] a, int localSize, int[] result, int[] summe) {
    this.a = a;
    this.result = result;
    this.summe = summe;
    scratch = new int[localSize];
  }

  @Override
  public void run() {
    if (this.step == 1) {
      step1();
    } else {
      step2();
    }
  }

  public void setStep(int step) {
    this.step = step;
  }

  private void step1() {
    int globalId = getGlobalId();
    int localId = getLocalId();
    int globalSize = getGlobalSize();

    int sum = 0;

    for (int i = globalId; i < a.length; i += globalSize) {
      sum += a[i];
    }

    scratch[localId] = sum;
    localBarrier();

    for (int s = getLocalSize() / 2; s > 0; s /= 2) {
      if (localId < s) {
        scratch[localId] += scratch[localId + s];
      }
      localBarrier();
    }

    if (localId == 0) {
      result[getGroupId()] = scratch[0];
    }
  }

  // Verarbeitung der einzelnen Elemente in result[]
  // Erst wieder einmal sequenziell, dann parallel
  private void step2() {
    int localId = getLocalId();
    int localSize = getLocalSize();
    scratch[localId] = result[localId] + result[localId + localSize];
    localBarrier(); // Wait for all to finish (locally)

    for (int s = localSize / 2; s > 0; s /= 2) {
      if (localId < s) {
        scratch[localId] += scratch[localId + s];
      }
      localBarrier();
    }
    if (localId == 0) {
      summe[0] = scratch[0];
    }
  }

}
